#!/usr/bin/env python

from __future__ import print_function
from __future__ import absolute_import
import argparse
import logging
import time

from zhue.model.bridge import Bridge
from yaml import load as yaml_load
import requests


def parse_args():
    ON_OFF_CHOICE = ['on', 'off']
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--username',
                        help='Username', required=True)
    parser.add_argument('-b', '--bridge',
                        help='Hostname or IP of the Hue bridge',
                        default=None, required=False)
    subparsers = parser.add_subparsers(dest='action', help='Action')
    light_parser = subparsers.add_parser('light')
    light_parser.add_argument('NAME')
    light_parser.add_argument('STATE', choices=ON_OFF_CHOICE)
    sensors_parser = subparsers.add_parser('sensors')
    sensor_parser = subparsers.add_parser('sensor')
    sensor_parser.add_argument('NAME')
    sensor_parser.add_argument('STATE', choices=ON_OFF_CHOICE, nargs='?')
    switch_parser = subparsers.add_parser('switch')
    switch_parser.add_argument('NAME')
    switch_parser.add_argument('-w', '--wait', action='store_true',
                               default=False)
    switch_parser.add_argument('-S', '--secrets-file',
                               required=False, type=argparse.FileType('r'),
                               help='Home Assistant secrets file')
    switch_parser.add_argument('-H', '--home-assistant',
                               required=False,
                               help='Home Assistant URL')
    switch_parser.add_argument('-P', '--home-assistant-password',
                               required=False,
                               help='Home Assistant Password')
    temperature_parser = subparsers.add_parser('temperature')
    temperature_parser.add_argument('NAME')
    light_level_parser = subparsers.add_parser('light-level')
    light_level_parser.add_argument('NAME')
    motion_parser = subparsers.add_parser('motion')
    motion_parser.add_argument('NAME')
    battery_parser = subparsers.add_parser('battery')
    battery_parser.add_argument('NAME')
    return parser.parse_args()


def hass_event(url, api_key, switches, event_name='hue_dimmer_switch_pressed'):
    last_updated = {}
    while True:
        for switch in switches:
            last_updated[switch.name] = switch.last_updated
            switch.update()
            if switch.last_updated != last_updated[switch.name]:
                button, click_type = switch.friendly_button_event
                button, click_type = switch.friendly_button_event
                print('Name:', switch.name, '\nButton:', button,
                      '\nEvent:', click_type, '\n')
                if url:
                    res = requests.post(
                        '{}/api/events/{}'.format(url, event_name),
                        headers={'X-HA-Access': api_key},
                        json={'button_name': button,
                              'click_type': click_type}
                    )
                    res.raise_for_status()
                    print(res.json(), '\n')
                last_updated[switch.name] = switch.last_updated
            time.sleep(1)


def main():
    args = parse_args()
    if args.bridge:
        bridge = Bridge(args.bridge, username=args.username)
    else:
        bridges = Bridge.discover_nupnp(username=args.username)
        bridge = bridges[0]
    if args.action == 'lights':
        print(bridge.lights)
    elif args.action == 'light':
        light = bridge.light(args.NAME)
        if args.STATE == 'on':
            light.on()
        else:
            light.off()
    elif args.action == 'sensor':
        sensor = bridge.sensor(args.NAME)
        if args.STATE == 'on':
            print(sensor.enable())
        elif args.STATE == 'off':
            print(sensor.disable())
        else:
            print(sensor._json)
    elif args.action == 'sensors':
        print(bridge.sensors)
    elif args.action == 'switch':
        if args.NAME.lower() == 'all':
            switches = bridge.switches
        else:
            switches = [bridge.switch(args.NAME)]
        if args.wait:
            if args.secrets_file:
                secrets_file = yaml_load(args.secrets_file)
                hass_url = secrets_file.get('base_url')
                hass_key = secrets_file.get('http_password')
            else:
                hass_url = args.home_assistant
                hass_key = args.home_assistant_password
            hass_event(hass_url, hass_key, switches)
        else:
            for switch in switches:
                button, click_type = switch.friendly_button_event
                print('Name:', switch.name, '\nButton:', button,
                      '\nEvent:', click_type)
    elif args.action == 'battery':
        device = bridge.device(args.NAME)
        print(device.battery)
    elif args.action == 'temperature':
        device = bridge.device(args.NAME)
        print(device.temperature)
    elif args.action == 'light-level':
        device = bridge.device(args.NAME)
        print(device.light_level)
    elif args.action == 'motion':
        device = bridge.device(args.NAME)
        print(device.presence)


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import unicode_literals
from .basemodel import HueLLDevice


class Rule(HueLLDevice):
    def __init__(self, bridge, hue_id, json):
        super(Rule, self).__init__(bridge, 'rules', hue_id, json)

    @staticmethod
    def new(bridge, *args, **kwargs):
        raise NotImplementedError()

    @property
    def owner(self):
        return self._bridge.user(username=self._json['owner'])

    @property
    def lights(self):
        l = []
        for l_id in self._json['lights']:
            l.append(self._bridge.light(hue_id=l_id))
        return l

    @property
    def locked(self):
        return self._json['locked']

    @property
    def picture(self):
        return self._json['picture']

    @property
    def recycle(self):
        return self._json['recycle']

    @property
    def version(self):
        return self._json['version']

    @property
    def status(self):
        return self._json['status']

    @property
    def created(self):
        return self._json['created']

    @property
    def lasttriggered(self):
        return self._json['lasttriggered']

    @property
    def timestriggered(self):
        return self._json['timestriggered']

    @property
    def conditions(self):
        return self._json['conditions']

    @property
    def actions(self):
        return self._json['actions']



# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals
from .basemodel import (HueLLDevice, HueJsonObject)
from .api_response import HueSuccessResponse
import re


class Schedule(HueLLDevice):
    def __init__(self, bridge, hue_id, json):
        super(Schedule, self).__init__(bridge, 'schedules', hue_id, json)

    @staticmethod
    def new(bridge, command, localtime, status='enabled', name='',
            description='', autodelete=True, recycle=True):
        if isinstance(command, ScheduledCommand):
            cmd = command._json
        else:
            cmd = command
        if isinstance(localtime, ScheduleTime):
            ltime = localtime.timestr
        else:
            ltime = localtime
        response = bridge._request(
            url='{}/schedules'.format(bridge.API),
            method='POST',
            data={
                'command': cmd,
                'name': name,
                'description': description,
                'status': status,
                'localtime': ltime,
                # 'autodelete': autodelete, # FIXME
                'recycle': recycle
            }
        )
        if type(response) is HueSuccessResponse:
            return bridge.schedule(hue_id=response._json['id'])
        return response

    @property
    def description(self):
        return self._json['description']

    @property
    def created(self):
        return self._json['created']

    @property
    def time(self):
        return ScheduleTime(self._json['time'])

    @property
    def localtime(self):
        return ScheduleTime(self._json['localtime'])

    @localtime.setter
    def localtime(self, value):
        if isinstance(value, ScheduleTime):
            val = value.timestr
        else:
            val = value
        self._request(method='PUT', data={'localtime': val})
        self.update()
        # self._json['localtime'] = val

    @property
    def command(self):
        return ScheduledCommand(self._json['command'])

    @property
    def device(self):
        return self._bridge.from_address(self.command.address)

    @property
    def status(self):
        return self._json['status']

    @property
    def recycle(self):
        return self._json['recycle']


class ScheduleTime(object):
    def __init__(self, timestr):
        self.timestr = timestr

    @property
    def recurring(self):
        '''
        # Mo Tu We Th Fr Sa Su
        # 64 32 16  8  4  2  1

                         Mo   Tu   We  Th  Fr  Sa  Su
        Weekdays:  124 = 64 + 32 + 16 + 8 + 4

        Mo-Th Sun: 121 = 64 + 32 + 16 + 8 + 0 + 0 + 1
        '''
        return re.match(
            r'W\d{3}/T\d{2}:\d{2}:\d{2}(A\d{2}:\d{2}\d{2})?',
            self.timestr
        ) is not None

    # @property
    # def randomized(self):
    #     return re.match(
    #         r'(W\d{3}/T)?\d{2}:\d{2}:\d{2}(A\d{2}:\d{2}\d{2})?',
    #         self.timestr
    #     ) is not None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '<ScheduleTime: {}>'.format(self.timestr)


class RecurringTime(ScheduleTime):
    pass


class ScheduledCommand(HueJsonObject):
    @property
    def method(self):
        return self._json['method']

    @property
    def body(self):
        return self._json['body']

    @property
    def address(self):
        '''
        Override the address property because if does not contain the
        bridge hostname
        '''
        return self._json['address']



#!/usr/bin/env python
# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import re
import requests
from simplejson.decoder import JSONDecodeError

from .api_response import (HueApiResponse, HueErrorResponse)
from .config import BridgeConfig
from .device import HueDevice
from .group import (Group, MasterGroup)
from .light import Light
from .rule import Rule
from .scene import Scene
from .schedule import Schedule
from .sensor import (LightLevelSensor, SwitchSensor,
                     TemperatureSensor, PresenceSensor)
from .sensor import factory as sensorfactory


_LOGGER = logging.getLogger(__name__)


class HueError(Exception):
    pass


class Bridge(object):
    def __init__(self, hostname, username=None, port=80):
        self.hostname = hostname
        self.port = port
        self._username = username
        self.__contruct_api_url()

    def from_address(self, address):
        endpoint_regex = re.compile(r'/api/([^/]+)/([^/]+)/([^/]+)(?:/(.+))?')
        match = re.match(endpoint_regex, address)
        assert match, 'Invalid address'
        endpoint = match.group(2)
        hue_id = match.group(3)
        if endpoint == 'sensors':
            return self.sensor(hue_id=hue_id)
        elif endpoint == 'lamps':
            return self.light(hue_id=hue_id)
        elif endpoint == 'groups':
            return self.group(hue_id=hue_id)
        elif endpoint == 'schedules':
            return self.schedule(hue_id=hue_id)
        raise HueError('Unsupported enpoint')

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value
        self.__contruct_api_url()

    def __contruct_api_url(self):
        self.API = 'http://{}:{}/api'.format(
            self.hostname,
            self.port
        )
        if self.username:
            self.API += '/{}'.format(self.username)
        _LOGGER.info('Update API URL to: {}'.format(self.API))

    def _request(self, url, method='GET', data=None):
        res = requests.request(url=url, method=method, json=data)
        logging.debug('Request data: {}'.format(data))
        if not res.ok:
            res.raise_for_status()
        try:
            jr = res.json()
            _LOGGER.debug('JSON Response: {}'.format(jr))
            response = HueApiResponse.factory(jr)
            if type(response) is HueErrorResponse:
                raise HueError(response.description)
            return response
        except JSONDecodeError:
            _LOGGER.error(
                'Failed to decode JSON from response: {}'.format(res.text)
            )

    @staticmethod
    def from_url(url, username=None):
        from urlparse import urlparse
        res = urlparse(url)
        return Bridge(hostname=res.hostname, port=res.port, username=username)

    @staticmethod
    def discover(username=None, guess=True, fast=True):
        nupnp = Bridge.discover_nupnp(username=username)
        if not fast or not nupnp:
            upnp = Bridge.discover_upnp(username=username)
        else:
            _LOGGER.info('Skip UPNP discovery')
            upnp = []
        bridges = list(set(upnp + nupnp))
        return bridges[0] if guess and bridges else bridges

    @staticmethod
    def discover_nupnp(username=None):
        res = requests.get('https://www.meethue.com/api/nupnp')
        return [Bridge(x['internalipaddress'], username) for x in res.json()]

    @staticmethod
    def discover_upnp(username=None):
        from netdisco.ssdp import scan

        PHILIPS = 'Royal Philips Electronics'
        BRIDGE_MODELS = ['929000226503', 'BSB002']

        hue_bridges = []
        res = scan()

        for h in res:
            _LOGGER.info('Check {}'.format(h))
            device_info = h.description.get('device', None)
            if device_info:
                manufacturer = device_info.get('manufacturer', None)
                model = device_info.get('modelNumber', None)
                if manufacturer == PHILIPS and model in BRIDGE_MODELS:
                    url = h.description['URLBase']
                    if url not in hue_bridges:
                        hue_bridges.append(url)
            else:
                _LOGGER.error('Missing device info')
        return [Bridge.from_url(x, username) for x in hue_bridges]

    def delete_user(self, username):
        url = '{}/config/whitelist/{}'.format(self.API, username)
        return self._request(method='DELETE', url=url)

    def delete_light(self, hue_id):
        url = '{}/lights/{}'.format(self.API, hue_id)
        return self._request(method='DELETE', url=url)

    def get_full_state(self):
        return self._request(method='GET', url=self.API)

    def _property(self, prop_url):
        url = '{}/{}'.format(self.API, prop_url)
        res = self._request(url)
        return res._json

    @property
    def config(self):
        return BridgeConfig(self, self._property('config'))

    @property
    def users(self):
        return self.config.users

    @property
    def lights(self):
        l = []
        for i, v in self._property('lights').items():
            l.append(Light(self, i, v))
        return l

    @property
    def groups(self):
        l = []
        for i, v in self._property('groups').items():
            l.append(Group(self, i, v))
        return l

    @property
    def schedules(self):
        l = []
        for i, v in self._property('schedules').items():
            l.append(Schedule(self, i, v))
        return l

    @property
    def sensors(self):
        s = []
        for i, v in self._property('sensors').items():
            s.append(sensorfactory(self, i, v))
        return s

    @property
    def scenes(self):
        s = []
        for i, v in self._property('scenes').items():
            s.append(Scene(self, i, v))
        return s

    @property
    def devices(self):
        devs = []
        sensors_grouped = self.__regroup_sensors()
        for d in filter(None, sensors_grouped):
            devs.append(HueDevice.factory(sensors_grouped[d]))
        devs += self.lights
        return devs

    @property
    def rules(self):
        r = []
        for i, v in self._property('scenes').items():
            r.append(Rule(self, i, v))
        return r

    def __regroup_sensors(self):
        d = {}
        for s in self.sensors:
            mac_addr = None
            try:
                mac_addr = s.mac_address
            except:
                pass
            if mac_addr in d:
                d[mac_addr].append(s)
            else:
                d[mac_addr] = [s]
        return d

    def __get_sensors_by_type(self, sensor_type):
        return [x for x in self.sensors if isinstance(x, sensor_type)]

    @property
    def temperature_sensors(self):
        return self.__get_sensors_by_type(TemperatureSensor)

    @property
    def light_level_sensors(self):
        return self.__get_sensors_by_type(LightLevelSensor)

    @property
    def presence_sensors(self):
        return self.__get_sensors_by_type(PresenceSensor)

    @property
    def switches(self):
        return self.__get_sensors_by_type(SwitchSensor)

    def __get_hue_objects_by_type(self, device_type):
        if device_type == 'sensor':
            return self.sensors
        elif device_type == 'switch':
            return self.switches
        elif device_type == 'light':
            return self.lights
        elif device_type == 'group':
            return self.groups
        elif device_type == 'schedule':
            return self.schedules
        elif device_type == 'scene':
            return self.scenes
        elif device_type == 'user':
            return self.users
        elif device_type == 'device':
            return self.devices
        elif device_type == 'rule':
            return self.rules
        else:
            _LOGGER.error('Unknown device type')

    def __get_hue_object(self, device_type, name=None, hue_id=None,
                         exact=False):
        assert name or hue_id, 'Name or Hue ID must be provided'
        for d in self.__get_hue_objects_by_type(device_type):
            if hue_id and str(d.hue_id) == str(hue_id):
                return d
            if name:
                if exact:
                    if d.name == name:
                        return d
                else:
                    if re.match('.*{}.*'.format(name), d.name, re.IGNORECASE):
                        return d
        raise HueError('No matching item was found')

    def group(self, name=None, hue_id=None, exact=False):
        if hue_id == 0:
            return MasterGroup(self)
        return self.__get_hue_object('group', name, hue_id, exact)

    def light(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('light', name, hue_id, exact)

    def sensor(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('sensor', name, hue_id, exact)

    def switch(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('switch', name, hue_id, exact)

    def schedule(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('schedule', name, hue_id, exact)

    def scene(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('scene', name, hue_id, exact)

    def user(self, name=None, username=None, exact=False):
        return self.__get_hue_object('user', name, username, exact)

    def device(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('device', name, hue_id, exact)

    def rule(self, name=None, hue_id=None, exact=False):
        return self.__get_hue_object('rule', name, hue_id, exact)

    # Hue object discovery
    def __find_new(self, hueobjecttype):
        '''
        Starts a search for new Hue objects
        '''
        assert hueobjecttype in ['lights', 'sensors'], \
            'Unsupported object type {}'.format(hueobjecttype)
        url = '{}/{}'.format(self.API, hueobjecttype)
        return self._request(
            method='POST',
            url=url
        )

    def __get_new(self, hueobjecttype):
        '''
        Get a list of newly found Hue object
        '''
        assert hueobjecttype in ['lights', 'sensors'], \
            'Unsupported object type {}'.format(hueobjecttype)
        url = '{}/{}/new'.format(self.API, hueobjecttype)
        return self._request(url=url)

    def find_new_lights(self):
        return self.__find_new('lights')

    def new_lights(self):
        return self.__get_new('lights')

    def find_new_sensors(self):
        return self.__find_new('sensors')

    def new_sensors(self):
        return self.__get_new('sensors')

    # All on/off
    def all_off(self):
        return MasterGroup(self).off()

    def all_on(self):
        return MasterGroup(self).on()

    # Factory methods. Create new objects
    def create_user(self, devicetype='zhue.py#user'):
        url = 'http://{}:{}/api'.format(self.hostname, self.port)
        data = {'devicetype': devicetype}
        res = self._request(method='POST', url=url, data=data)
        self.username = res._json['username']
        return res

    def create_schedule(self, *args, **kwargs):
        return Schedule.new(self, *args, **kwargs)

    def create_rule(self, *args, **kwargs):
        return Rule.new(self, *args, **kwargs)

    # Software update
    def update_check(self):
        return



from __future__ import absolute_import
from __future__ import unicode_literals
from .basemodel import (BaseGroup, LightDeviceState)


class Group(BaseGroup):
    def __init__(self, bridge, hue_id, json):
        super(Group, self).__init__(bridge, 'groups', hue_id, json)

    @property
    def name(self):
        if self._json:
            return super(Group, self).name
        elif self.hue_id == 0:
            return '# Master group'

    @property
    def state(self):
        return GroupState(
            dict(self._json['state'].items() + self._json['action'].items())
        )

    @property
    def all_on(self):
        return self.state.all_on

    @property
    def any_on(self):
        return self.state.any_on

    def _set_state(self, data):
        # Groups have a different endpoint for the "state" (/action)
        url = '{}/action'.format(self.API)
        res = self._request(
            method='PUT',
            url=url,
            data=data
        )
        self.update()
        return res

    def set_scene(self, scene):
        return self._set_state({'scene': scene.hue_id})


class GroupState(LightDeviceState):
    @property
    def all_on(self):
        return self._json['all_on']

    @property
    def any_on(self):
        return self._json['any_on']


class MasterGroup(Group):
    def __init__(self, bridge):
        super(MasterGroup, self).__init__(bridge=bridge, hue_id=0, json=None)

    @property
    def lights(self):
        return self._bridge.ligths



from __future__ import absolute_import
from __future__ import unicode_literals
from .sensor import (PresenceSensor, TemperatureSensor, LightLevelSensor)
from .basemodel import HueLLDevice


class HueDevice(HueLLDevice):
    @staticmethod
    def factory(devices):
        # If devices only holds a single sensor, use it
        if len(devices) == 1:
            return devices[0]
        presence_sensor = False
        temp_sensor = False
        light_sensor = False
        for d in devices:
            if isinstance(d, PresenceSensor):
                presence_sensor = True
            elif isinstance(d, TemperatureSensor):
                temp_sensor = True
            elif isinstance(d, LightLevelSensor):
                light_sensor = True
        if presence_sensor and temp_sensor and light_sensor:
            return HueMotionDetectorDevice(devices)
        raise RuntimeError('Unsupported device')


class HueMultiSensorDevice(HueDevice):
    def __init__(self, sensors):
        self.sensors = sensors

    def _get_sensor_by_type(self, sensor_type):
        for s in self.sensors:
            if isinstance(s, sensor_type):
                return s

    @property
    def mac_address(self):
        return self.sensors[0].mac_address


class HueMotionDetectorDevice(HueMultiSensorDevice):
    @property
    def presence_sensor(self):
        return self._get_sensor_by_type(PresenceSensor)

    @property
    def temperature_sensor(self):
        return self._get_sensor_by_type(TemperatureSensor)

    @property
    def light_level_sensor(self):
        return self._get_sensor_by_type(LightLevelSensor)

    @property
    def name(self):
        return self.presence_sensor.name

    @property
    def temperature(self):
        return self.temperature_sensor.temperature

    @property
    def light_level(self):
        return self.light_level_sensor.light_level

    @property
    def presence(self):
        return self.presence_sensor.presence

    @property
    def battery(self):
        return self.sensors[0].battery

    @property
    def hue_id(self):
        return self.sensors[0].hue_id



from __future__ import absolute_import
from __future__ import unicode_literals
from .basemodel import (LightDevice, LightDeviceState)


class Light(LightDevice):
    def __init__(self, bridge, hue_id, json):
        super(Light, self).__init__(bridge, 'lights', hue_id, json)

    @property
    def state(self):
        return LightDeviceState(self._json['state'])

    def delete(self):
        return self._bridge.delete_light(self.hue_id)

    def __str__(self):
        return '<Light: {}{}>'.format(self.name, ' (*)' if self.is_on else '')



from __future__ import absolute_import
from __future__ import unicode_literals
from .basemodel import HueBaseObject
from .user import User


class BridgeConfig(HueBaseObject):
    def __init__(self, bridge, json):
        super(BridgeConfig, self).__init__(bridge, 'config', json)

    # Versions
    @property
    def api_version(self):
        return self._json['apiversion']

    @property
    def version(self):
        return self._json['swversion']

    @property
    def bridge_id(self):
        return self._json['bridgeid']

    # Network stuff
    @property
    def dhcp(self):
        return self._json['dhcp']

    @property
    def gateway(self):
        return self._json['gateway']

    @property
    def mac(self):
        return self._json['mac']

    @property
    def netmask(self):
        return self._json['netmask']

    @property
    def zigbeechannel(self):
        return self._json['zigbeechannel']

    @property
    def factorynew(self):
        return self._json['factorynew']

    @property
    def timezone(self):
        return self._json['timezone']

    @property
    def localtime(self):
        return self._json['localtime']

    @property
    def utc(self):
        return self._json['UTC']

    @property
    def users(self):
        u = []
        for k, v in self._json['whitelist'].items():
            u.append(User(username=k, json=v))
        return u

    def update_check(self):
        res = self._request(data={"swupdate": {"checkforupdate":True}})
        self.update()
        return res



from __future__ import unicode_literals
from __future__ import print_function
import re


uuid_mac_regex = re.compile(
    r'(([0-9a-f]{2}[:-]){6}([0-9a-f]{2})):(.+)'
)


class HueJsonObject(object):
    def __init__(self, json):
        self._json = json


class NamedHueJsonObject(HueJsonObject):
    @property
    def name(self):
        return self._json['name']


class HueBaseObject(HueJsonObject):
    def __init__(self, bridge, api_endpoint, json):
        super(HueBaseObject, self).__init__(json)
        self._bridge = bridge
        self.api_endpoint = api_endpoint
        self.API = '{}/{}'.format(
            self._bridge.API,
            self.api_endpoint
        )

    # Shortcut function
    def _request(self, *args, **kwargs):
        if 'url' not in kwargs:
            return self._bridge._request(url=self.API, *args, **kwargs)
        return self._bridge._request(*args, **kwargs)

    def update(self):
        '''
        Update our object's data
        '''
        self._json = self._request(
            method='GET',
            url=self.API
        )._json


class HueObject(HueBaseObject):
    def __init__(self, bridge, api_endpoint, hue_id, json):
        super(HueObject, self).__init__(bridge, api_endpoint, json)
        self.hue_id = hue_id
        self.API = '{}/{}/{}'.format(
            self._bridge.API,
            self.api_endpoint,
            self.hue_id
        )

    def delete(self):
        return self._request(
            method='DELETE',
            url=self.API
        )

    @property
    def address(self):
        '''
        Return the address of this "object", minus the scheme, hostname
        and port of the bridge
        '''
        return self.API.replace(
            'http://{}:{}'.format(
                self._bridge.hostname,
                self._bridge.port
            ), ''
        )

    # Shortcut function
    def _request(self, *args, **kwargs):
        if 'url' not in kwargs:
            return self._bridge._request(url=self.API, *args, **kwargs)
        return self._bridge._request(*args, **kwargs)

    def update(self):
        '''
        Update our object's data
        '''
        self._json = self._request(
            method='GET',
            url=self.API
        )._json


class HueLLDevice(HueObject):
    def _set_state(self, data):
        url = '{}/state'.format(self.API)
        res = self._request(
            method='PUT',
            url=url,
            data=data
        )
        self.update()
        return res

    @property
    def manufacturer(self):
        return self._json['manufacturername']

    @property
    def model(self):
        return self._json['modelid']

    @property
    def name(self):
        return self._json['name']

    @name.setter
    def name(self, value):
        self._request(
            method='PUT',
            data={'name': value}
        )
        self.update()

    @property
    def version(self):
        return self._json['swversion']

    @property
    def uuid(self):
        return self._json.get('uniqueid', None)

    def __decompose_uuid(self):
        if self.uuid:
            m = re.match(uuid_mac_regex, self.uuid)
            if m:
                return m.group(1), m.group(4)
        return None, None

    @property
    def mac_address(self):
        mac, _ = self.__decompose_uuid()
        return mac

    @property
    def device_id(self):
        _, dev_id = self.__decompose_uuid()
        return dev_id

    @property
    def type(self):
        return self._json['type']

    @property
    def state(self):
        return self._json['state']

    # @property
    # def config(self):
    #     return self._json['config']

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return '<{}: {}>'.format(type(self).__name__, self.name)


class LightDevice(HueLLDevice):
    def __on_off(self, state):
        assert state in ['on', 'off'], 'Unknown state: {}'.format(state)
        return self._set_state(data={'on': state == 'on'})

    def on(self):
        return self.__on_off('on')

    def off(self):
        return self.__on_off('off')

    @property
    def is_on(self):
        return self.state.on

    def toggle(self):
        return self.off() if self.is_on else self.on()

    def delete(self):
        return self._request(method='DELETE')

    def alert(self, effect='select'):
        assert effect in ['select', 'lselect', 'none'], \
            'Unknown alert effect: {}'.format(effect)
        return self._set_state(data={'alert': effect})

    def alert_stop(self):
        return self.alert('none')

    @property
    def brightness(self):
        return self.state.brightness

    @brightness.setter
    def brightness(self, value):
        assert value in xrange(0, 256), 'Unsupported brigthness value'
        return self._set_state(data={'bri': value})

    @property
    def saturation(self):
        return self.state.saturation

    @saturation.setter
    def saturation(self, value):
        assert value in xrange(0, 256), 'Unsupported saturation value'
        return self._set_state(data={'sat': value})

    @property
    def hue(self):
        return self.state.hue

    @hue.setter
    def hue(self, value):
        assert value in xrange(0, 65536), 'Unsupported hue value'
        return self._set_state(data={'hue': value})

    @property
    def xy(self):
        return self.state.xy

    @xy.setter
    def xy(self, value):
        assert type(value) is list, 'Unsupported xy value'
        return self._set_state(data={'xy': value})

    @property
    def color_temperature(self):
        return self.state.color_temperature

    @color_temperature.setter
    def color_temperature(self, value):
        assert value in xrange(0, 65536), 'Unsupported color_temperature value'
        return self._set_state(data={'ct': value})


class LightDeviceState(HueJsonObject):
    @property
    def alert(self):
        return self._json['alert']

    @property
    def on(self):
        return self._json['on']

    @property
    def brightness(self):
        return self._json['bri']

    @property
    def colormode(self):
        return self._json['colormode']

    @property
    def color_temperature(self):
        return self._json['ct']

    @property
    def effect(self):
        return self._json['effect']

    @property
    def hue(self):
        return self._json['hue']

    @property
    def reachable(self):
        return self._json['reachable']

    @property
    def saturation(self):
        return self._json['sat']

    @property
    def xy(self):
        return self._json['xy']


class BaseGroup(LightDevice):
    @property
    def lights(self):
        l = []
        for l_id in self._json['lights']:
            l.append(self._bridge.light(hue_id=l_id))
        return l
